#include <TimerOne.h>


/*
Generuje PWM signal s interaktivne zadanou stridou a frekvenci. Pro zmenu stridy 
poslete na konzoli vstup ve tvaru s+NOVA_STRIDA (napr. d50 nastavi stridu 50%). 
Pro zmenu frekvence zadejte f+NOVA_FREKVENCE (napr. f1000 nastavi frekvenci 
na 1000 Hz.) Pro zmenu doby prumerovani vzorku pouzijte a+NOVA_HODNOTA (napr. 
a100 nastavi dobu vzorkovani na 100 ms)
Arduino zaroven meri napeti na pinech A1,A2 a A3 a vystup posila na konzoli. Pro
zobrazeni pouzijte program Putty s Baud rate 9600.
*/

int strida=50;
long perioda=1000;
int prumerovani=500;
char mode='b';
const int graph_posr=6;
const int graph_posc=5;

void setup() {
  Serial.begin(115200);
  Timer1.initialize(1000);
  Timer1.pwm(9,512);
  Timer1.start();
  print_init();
  tisk_nastaveni(strida,perioda, prumerovani);
  help();
}


const int hist_len=30;
bool new_visual;
bool use_reprint=false;
bool pause=false;

void loop() {
  if (!pause) measure(prumerovani);
  
  if (Serial.available() > 0) {   //vstup od uzivatele
    char c=Serial.read();
    if (c=='d'){
      strida=Serial.parseInt();
      if (strida<0) strida=0;
      if (strida>100) strida=100;
      set_PWM(perioda, strida);
      if (mode!='l' && mode!='i') tisk_nastaveni(strida,perioda, prumerovani);
    } else if (c=='f') {    //insert frequency
      perioda=Serial.parseInt(); //actually frequency at this moment
      if (perioda==0) {
        perioda=1000000;
      }else{
        perioda=1000000/perioda;
      }
      if (perioda<100) perioda=100;
      if (perioda>1000000) perioda=1000000;
      set_PWM(perioda, strida);
      if (mode!='l' && mode!='i') tisk_nastaveni(strida,perioda, prumerovani);
    } else if (c=='a'){
      prumerovani=Serial.parseInt();
      if (prumerovani<20) prumerovani=20;
      if (prumerovani>1000) prumerovani=1000;
      //if(prumerovani<200 && (mode=='0' || mode=='1' || mode=='2')) prumerovani=200;
      if (mode!='l' && mode!='i') tisk_nastaveni(strida,perioda, prumerovani);
    } else if (c=='i'){
      print_info();
      mode='i';
    } else if (c=='l'){ //logging mode
      mode='l';
      print_init();
      pause=true;
      Serial.print(char(27));
      Serial.print("[1;1H");
      Serial.print("Press 'p' to start/pause measuring.\n\r");
    } else if (c=='r'){ //reset or enter basic mode
      mode='b';
      pause=false;
      print_init();
      tisk_nastaveni(strida,perioda, prumerovani);
      help();
      use_reprint=false;
    } else if (c=='p'){ //reset or enter basic mode
      pause=!pause;
    } else if (c=='v'){ //visual mode
      pause=false;
      int vnode=Serial.parseInt();
      if (vnode==0) {
        mode='0';
      }else if (vnode==1) {
        mode='1';
      }else if (vnode==2) {
        mode='2';
      } else{
        while(Serial.available() > 0) Serial.read();
        return;
      }
      print_init();
      tisk_nastaveni(strida,perioda, prumerovani);
      new_visual=true;
      //if(prumerovani<200) prumerovani=200;
    }
    while(Serial.available() > 0) Serial.read();
  }  
}

//unsigned long timer2_ms=0;
unsigned long timer_ms;
unsigned long timer_us;
long a0, a1, a2;
long a0_prumer, a1_prumer, a2_prumer;

void measure(int prumerovani){
  //timer_ms=millis();
  a0_prumer=a1_prumer=a2_prumer=0;
  for(int i=0;i<prumerovani/20;i++) {
    a0=a1=a2=0;
    for (int j=0;j<50;j++) {
      timer_us=micros();
      a0+=analogRead(A0);
      //delayMicroseconds(900);
      a1+=analogRead(A1);
      //delayMicroseconds(900);
      a2+=analogRead(A2);
      if (400-(micros()-timer_us)>100){
        delayMicroseconds(100);
      }else{
        delayMicroseconds(400-(micros()-timer_us)-5);
      }
    }
    a0_prumer+=a0/50;
    a1_prumer+=a1/50;
    a2_prumer+=a2/50;
  }

  
  a0_prumer=3125*a0_prumer/(32*prumerovani);    //5000*a0_prumer/(10*1024)
  a1_prumer=3125*a1_prumer/(32*prumerovani);
  a2_prumer=3125*a2_prumer/(32*prumerovani);

  
  
  if (mode=='l'){
    Serial.print(a0_prumer);
    Serial.print("\t");
    Serial.print(a1_prumer);
    Serial.print("\t");
    Serial.print(a2_prumer);
    Serial.print("\n\r");
  } else if (mode=='0'){
    visual(a0_prumer);
    /*
    go2pos(20, 0);
    Serial.print(char(27));
    Serial.print("[K");
    Serial.print(millis()-timer2_ms);
    timer2_ms=millis();
    */
    visual_delay(25);       //delete if accuracy not needed 
  } else if (mode=='1'){
    visual(a1_prumer);
    visual_delay(25);       //delete if accuracy not needed 
  } else if (mode=='2'){
    visual(a2_prumer);
    visual_delay(25);       //delete if accuracy not needed 
  } else if (mode=='b'){
    Serial.print(char(27));
    Serial.print("[5;1H");
    if (use_reprint){
      reprint_values(a0_prumer, 6);
      reprint_values(a1_prumer, 7);
      reprint_values(a2_prumer, 8);
      reprint_values(a0_prumer-a1_prumer, 9);
      reprint_values(a1_prumer-a2_prumer, 10);
      reprint_values(a0_prumer-a2_prumer, 11);
    }else{
      Serial.print("\t=========================================\n\r");
      print_value(a0_prumer, "Sensor 0(A0)");
      print_value(a1_prumer, "Sensor 1(A1)");
      print_value(a2_prumer, "Sensor 2(A2)");
      print_value(a0_prumer-a1_prumer, "A0 - A1");
      print_value(a1_prumer-a2_prumer, "A1 - A2");
      print_value(a0_prumer-a2_prumer, "A0 - A2");
      use_reprint=true;
    }
    //if exact accuracy wanted use (sampling then will be averaging time + 5ms but stable):
    //visual_delay(5);   
  }
}

void visual_delay(int len){
  if ((prumerovani/20)*20+len-(millis()-timer_ms)>len+10){
     return;
  }else if ((prumerovani/20)*20+len-(millis()-timer_ms)>0){
     delay((prumerovani/20)*20+len-(millis()-timer_ms));
  }
  timer_ms=millis();
}

void visual_init(){
  go2pos(graph_posr-2, 0);
  Serial.print("\t\t\t   +25ms!!!\n\r");
  Serial.print(char(27));
  Serial.print("[K");
  Serial.print("Sensor ");
  Serial.print(mode);
  Serial.print(" [mV]");
  go2pos(graph_posr, 0);
  for (int i=10; i>0; i--){
    Serial.print(500*i);
    if (500*i<1000){
      Serial.print(" ");
    }
    Serial.print("|\n\r");
  }
  Serial.print("0   |\n\r");
  Serial.print("     ------------------------------\n\n\r");
  Serial.print("'f' + frequency in Hz (1-10000) (e.g. 'f100')\n\r");
  Serial.print("'d' + new duty cycle in % (0-100) (e.g. 'd50')\n\r");
  Serial.print("'a' + averaging period in ms (20-1000) (e.g. 'a100')\n\r");
  Serial.print("'r' to reset display size and go back to this mode\n\r");
  Serial.print("'p' to pause/unpause");
  //Serial.print("'r' to reset display size and go back to this mode\n\r");
}

int hist[hist_len];

void visual(long value){
  if (new_visual){
    for (int i=0; i<hist_len; i++){
      hist[i]=0;
    }
    visual_init();
    new_visual=false;
  }
  for (int i = 1; i < hist_len; i++){        
    hist[i-1]=hist[i];
  } 
  hist[hist_len-1]=value/500;
  go2pos(graph_posr, graph_posc);
  for (int val=1; val<hist_len; val++){
    if (hist[val-1]!=hist[val]){
      go2pos(graph_posr+10-hist[val-1], graph_posc+val);
      Serial.print(" ");
      go2pos(graph_posr+10-hist[val], graph_posc+val);
      Serial.print(".");
    }
  }
  /*
  go2pos(20, graph_posc);
  for (int i=0; i<20; i++){
    Serial.print(hist[i]);
    Serial.print(" ");
  }
  */
}


void go2pos(int row, int col){
  Serial.print(char(27));
  Serial.print("[");
  Serial.print(row);
  Serial.print(";");
  Serial.print(col);
  Serial.print("H");
}

  
void print_init(){
  Serial.print(char(27));
  Serial.print("[30;47m"); 
  Serial.print(char(27));
  Serial.print("[2J");
}

void tisk_nastaveni(int strida, long perioda, int prumerovani){
  Serial.print(char(27));
  Serial.print("[1;1H");
  Serial.print(char(27));
  Serial.print("[K");
  Serial.print("Settings:\n\r");
  Serial.print(char(27));
  Serial.print("[K");
  Serial.print("Frequency: ");
  Serial.print(1000000/perioda);
  Serial.print(" Hz\tDuty:");
  Serial.print(strida);
  Serial.print(" %\n\r");
  Serial.print(char(27));
  Serial.print("[K");
  Serial.print("Length of averaging period:");
  Serial.print(20*(prumerovani/20));
  Serial.print(" ms\n\r");
  Serial.print("\n\r");
  //Serial.print("\t=========================================\n\r");
}

void print_value(int value, char val_name[15]){
  Serial.print(char(27));
  Serial.print("[K");
  Serial.print("\t|\t");
  Serial.print(val_name);
  Serial.print(":\t");
  Serial.print(value);
  Serial.print(" mV\t\t|\n\r");
}

void reprint_values(int value, int pos){
  go2pos(pos,33);
  Serial.print(value);
  Serial.print(" mV   ");
}



void set_PWM(long perioda, int strida){
  Timer1.stop();
  Timer1.setPeriod(perioda);
  Timer1.setPwmDuty(9,map(strida,0,100,0,1023));
  Timer1.start();
}   

void help(){
  Serial.print(char(27));
  Serial.print("[12;1H");
  Serial.print("\t=========================================\n\r");
  Serial.print("\n\r");
  Serial.print(char(27));
  Serial.print("[K"); 
  Serial.print("Changing parameters:\n\r");
  Serial.print("'f' + frequency in Hz (1-10000) (e.g. 'f100')\n\r");
  Serial.print("'d' + new duty cycle in % (0-100) (e.g. 'd50')\n\r");
  Serial.print("'a' + averaging period in ms (20-1000) (e.g. 'a100')\n\r");
  Serial.print("'p' to pause/unpause");
  Serial.print("'r' to reset display size and go back to this mode\n\r");
  Serial.print("'l' to enter logging mode\n\r");
  Serial.print("'v' + channel number to enter visual mode\n\r");
  Serial.print("'i' for further information\n\r");
  Serial.print(char(27));
  Serial.print("[?25l");
}

void print_info(){
  Serial.print(char(27));
  Serial.print("[1;1H");
  Serial.print(char(27));
  Serial.print("[2J");
  Serial.println("PWM+Voltmeter");
  Serial.println("Simple PWM and Voltmeter program for Arduino UNO.");
  Serial.println("Uses pin ~9 as PWM source and pins A0-A2 for measuring.");
  Serial.println("PWM can generate signal with any frequency within");
  Serial.println("1-10000Hz wih step of 1Hz and duty cycle within");
  Serial.println("0-100% with step of 1%");
  Serial.println("3-channel voltmeter measures each of input pins every");
  Serial.println("400 us and prints average value for time selected by");
  Serial.println("user (any multiply of 20 ms to minimize effects of");
  Serial.println("50 Hz power supply). Arduino uses only one ADC for all");
  Serial.println("input pins and because it uses capacitor to stabilize");
  Serial.println("measurements last measured value will be detected on ");
  Serial.println("any unused pin.");
  Serial.println("Note: to close this message use 'r'");
  Serial.println("");
  Serial.println("Author: Jan Lindauer, K13138, CVUT FEL, Praha");
}
